import csv
import gzip
import argparse
import os
import io
from shutil import copyfile

parser = argparse.ArgumentParser(description = 'Script to restore reads from bin')
requiredargs = parser.add_argument_group('required arguments')
requiredargs.add_argument('--in', type = str, dest = 'input', help = 'Directory path of input data', required = True)
requiredargs.add_argument('--i7-map', type = str, dest = 'i7_map', help = 'Files containing mapping lists for i7', required = True, nargs = "+")
requiredargs.add_argument('--i5-map', type = str, dest = 'i5_map', help = 'Files containing mapping lists for i5', required = True, nargs = "+")
requiredargs.add_argument('--barcode-file', type = str, dest = 'barcode_file', help = 'File containing list of barcodes and their filename (all other files in the input directory are treated as undetermined/bin)', required = True)
parser.add_argument('--out', type = str, dest = 'output', help = 'Directory path of output (default = input directory/inline)')
parser.add_argument('--i5', type = str, dest = 'i5', default = [], help = 'List of all additional i5 barcodes used in the sequencing (by default only the barcodes defined in the barcode file are expected)', nargs = '*')
parser.add_argument('--i7', type = str, dest = 'i7', default = [], help = 'List of all additional i7 barcodes used in the sequencing (by default only the barcodes defined in the barcode file are expected)', nargs = '*')
parser.add_argument('--total-memory', type = int, dest = 'memory', help = 'Specifies the maximum allowed memory to be used for caching (in Bytes) (default = 1073741824)', default = 2**30)

#gets the next read (consisting of four lines)
def get_next_read(gzfile):
    read = []
    for line in gzfile:
        read.append(line.decode('utf-8'))
        if(len(read) == 4):
            barcodes = read[0][read[0].rfind(':') + 1:].strip().split('+')
            yield (barcodes, ''.join(read))
            read = []
    return

#loads all mappings
def load_maps(maps):
    mapping = dict()
    for path in maps:
        mapping[path] = dict()
        with open(path, "r") as gzfile:
            reader = csv.reader(gzfile, delimiter = '\t')
            for row in reader:
                mapping[path][row[0]] = row[1]
    return mapping

#reads the barcode file and gets all used barcodes and their corresponding files
def load_barcodes(barcode_file):
    with open(barcode_file, "r") as file:
        reader = csv.DictReader(file, delimiter = '\t')
        for row in reader:
            i5 = row.get('i5', None)
            i7 = row.get('i7')
            i5_set.add(i5)
            i7_set.add(i7)
            file_mapping[(i5, i7)] = row['file']
            file_set.add(row['file'])
    return (i5_set, i7_set, file_mapping, file_set)

args = parser.parse_args()

i5_set = set()
i7_set = set()

file_mapping = dict()
file_set = set()
file_handler = dict()

try:
    mapping_i7 = load_maps(args.i7_map)
    mapping_i5 = load_maps(args.i5_map)
    barcode_sets_i7 = {path: set(mapping_i7[path].values()) for path in mapping_i7}
    barcode_sets_i5 = {path: set(mapping_i5[path].values()) for path in mapping_i5}
    #handling special case of single indexing
    barcode_sets_i5[""] = set([None])
    mapping_i5[""] = set()
    #get all used indices and files
    (i5_set, i7_set, file_mapping, file_set) = load_barcodes(args.barcode_file)
    i5_set.update(args.i5)
    i7_set.update(args.i7)
    i5_key = None
    i7_key = None
    #find most appropriate barcode set
    for (key, barcodes) in sorted(barcode_sets_i7.items(), key = lambda x: len(x[1])):
        if(i7_set <= barcodes):
            i7_key = key
    for (key, barcodes) in sorted(barcode_sets_i5.items(), key = lambda x: len(x[1])):
        if(i5_set <= barcodes):
            i5_key = key
    if(i5_key is None or i7_key is None):
        print(i5_key, i7_key)
        print(i5_set, i7_set)
        print(barcode_sets_i7, barcode_sets_i5)
        raise Exception('Unknown set of barcodes defined')
    #find all fastq-files to process
    processing_paths = []
    for filename in os.listdir(args.input):
        path = os.path.join(args.input, filename)
        if(os.path.isfile(path) == True and filename not in file_set and path.endswith(".fastq.gz")):
            processing_paths.append(path)
    if(args.output is not None and os.path.exists(args.output) == False):
        os.makedirs(args.output)
    #Open all output files
    for key in file_mapping:
        out_file_path = os.path.join(args.output if args.output is not None else args.input, file_mapping[key])
        if(args.output is not None):
            copyfile(os.path.join(args.input, file_mapping[key]), out_file_path)
        file_handler[key] = io.BufferedWriter(gzip.open(out_file_path, mode = 'a', compresslevel = 4), 
                                              buffer_size = (args.memory // len(file_mapping)))
    saved_reads = 0
    #run through all files and all reads and map them to the correct barcodes
    for path in processing_paths:
        print("Correcting reads in %s"%(path, ))
        with gzip.open(path, 'r') as gz:
            for (barcodes, read) in get_next_read(gz):
                i7 = barcodes[0]
                i5 = barcodes[1] if(len(barcodes) == 2) else None
                key1 = mapping_i5[i5_key][i5] if i5 in mapping_i5[i5_key] else None
                key2 = mapping_i7[i7_key][i7] if i7 in mapping_i7[i7_key] else None
                key = (key1, key2)
                if(key in file_handler):
                    _ = file_handler[key].write(read.encode())
                    saved_reads += 1
    print("Successfully correcting errors in barcodes for %d reads"%(saved_reads))
finally:
    #close all files
    for key in file_handler:
        try:
            file_handler[key].close()
        except Exception:
            print('Warning: Problems with closing gzip file "%s"'%(file_handler[key].name,))

